import itertools
from typing import List, Optional

import netsquid as ns
import numpy as np
import pytest
from netsquid import QFormalism
from netsquid.nodes import Node
from netsquid_magic.state_delivery_sampler import (
    DeliverySample,
    DepolariseWithFailureStateSamplerFactory,
    PerfectStateSamplerFactory,
    StateDeliverySampler,
)

from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.lhi import LhiLinkInfo, LhiTopologyBuilder
from qoala.runtime.message import Message
from qoala.sim.build import build_qprocessor_from_topology
from qoala.sim.entdist.entdist import (
    DelayedSampler,
    EntDist,
    EntDistRequest,
    EprDeliverySample,
    JointRequest,
)
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.util.math import (
    B00_DENS,
    S10_DENS,
    TWO_MAX_MIXED,
    density_matrices_equal,
    has_multi_state,
)
from qoala.util.tests import netsquid_run


def create_n_nodes(n: int, num_qubits: int = 1) -> List[Node]:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits)
    nodes: List[Node] = []
    for i in range(n):
        qproc = build_qprocessor_from_topology(name=f"qproc_{i}", topology=topology)
        nodes.append(Node(name=f"node_{i}", qmemory=qproc))

    return nodes


def create_request(
    node1_id: int,
    node2_id: int,
    lpid: Optional[int] = 0,
    rpid: Optional[int] = 0,
) -> EntDistRequest:
    return EntDistRequest(
        local_node_id=node1_id,
        remote_node_id=node2_id,
        local_qubit_id=0,
        local_pid=lpid,
        remote_pid=rpid,
    )


def create_joint_request(
    node1_id: int,
    node2_id: int,
    node1_qubit_id: int = 0,
    node2_qubit_id: int = 0,
    node1_pid: int = 0,
    node2_pid: int = 0,
) -> JointRequest:
    return JointRequest(
        node1_id=node1_id,
        node2_id=node2_id,
        node1_qubit_id=node1_qubit_id,
        node2_qubit_id=node2_qubit_id,
        node1_pid=node1_pid,
        node2_pid=node2_pid,
    )


def create_entdist(nodes: List[Node]) -> EntDist:
    ehi_network = EhiNetworkInfo.only_nodes({node.ID: node.name for node in nodes})
    comp = EntDistComponent(ehi_network)
    return EntDist(nodes=nodes, ehi_network=ehi_network, comp=comp)


def test_add_sampler():
    alice, bob = create_n_nodes(2)

    entdist = create_entdist(nodes=[alice, bob])
    link_info = LhiLinkInfo.perfect(1000)
    entdist.add_sampler(alice.ID, bob.ID, link_info)

    assert len(entdist._samplers) == 1
    link = frozenset([alice.ID, bob.ID])
    sampler = entdist._samplers[link]
    assert type(sampler) == DelayedSampler
    assert type(sampler.sampler) == StateDeliverySampler
    assert sampler.delay == 1000


def test_add_sampler_many_nodes():
    n = 10
    nodes = create_n_nodes(n)

    entdist = create_entdist(nodes=nodes)
    link_info = LhiLinkInfo.perfect(1000)
    entdist.add_sampler(nodes[0].ID, nodes[1].ID, link_info)

    with pytest.raises(ValueError):
        entdist.add_sampler(nodes[0].ID, nodes[1].ID, link_info)
    with pytest.raises(ValueError):
        entdist.add_sampler(nodes[1].ID, nodes[0].ID, link_info)

    entdist.add_sampler(nodes[0].ID, nodes[2].ID, link_info)
    entdist.add_sampler(nodes[2].ID, nodes[9].ID, link_info)

    link = frozenset([nodes[0].ID, nodes[1].ID])
    assert entdist._samplers[link] == entdist.get_sampler(nodes[0].ID, nodes[1].ID)
    assert entdist._samplers[link] == entdist.get_sampler(nodes[1].ID, nodes[0].ID)

    with pytest.raises(ValueError):
        entdist.get_sampler(nodes[0].ID, nodes[9].ID)

    assert len(entdist._samplers) == 3
    for (i, j) in [(0, 1), (0, 2), (2, 9)]:
        link = frozenset([nodes[i].ID, nodes[j].ID])
        assert type(entdist._samplers[link].sampler) == StateDeliverySampler
        assert entdist._samplers[link].delay == 1000


def test_sample_perfect():
    sampler_factory = PerfectStateSamplerFactory()

    kwargs = {"cycle_time": 1000}
    sampler: StateDeliverySampler = sampler_factory.create_state_delivery_sampler(
        **kwargs
    )

    assert sampler._cycle_time == 1000
    assert sampler._success_probability == 1
    number_of_attempts = np.random.geometric(p=sampler._success_probability) - 1
    assert number_of_attempts == 0

    raw_sample: DeliverySample = sampler.sample()
    sample = EprDeliverySample.from_ns_magic_delivery_sample(raw_sample)

    expected_duration = 1000 * number_of_attempts  # = 0
    assert sample.duration == expected_duration
    assert density_matrices_equal(sample.state, B00_DENS)


def test_sample_depolar():
    sampler_factory = DepolariseWithFailureStateSamplerFactory()
    kwargs = {"cycle_time": 1000, "prob_max_mixed": 0.2, "prob_success": 1}
    sampler: StateDeliverySampler = sampler_factory.create_state_delivery_sampler(
        **kwargs
    )
    raw_sample: DeliverySample = sampler.sample()
    sample = EprDeliverySample.from_ns_magic_delivery_sample(raw_sample)

    expected = 0.2 * TWO_MAX_MIXED + 0.8 * B00_DENS

    assert density_matrices_equal(sample.state, expected)


def test_create_epr_pair_with_state():
    alice, bob = create_n_nodes(2)
    entdist = create_entdist(nodes=[alice, bob])

    q0, q1 = entdist.create_epr_pair_with_state(B00_DENS)
    assert has_multi_state([q0, q1], B00_DENS)

    q0, q1 = entdist.create_epr_pair_with_state(S10_DENS)
    assert has_multi_state([q0, q1], S10_DENS)


def test_deliver_perfect():
    alice, bob = create_n_nodes(2)

    entdist = create_entdist(nodes=[alice, bob])
    link_info = LhiLinkInfo.perfect(1000)
    entdist.add_sampler(alice.ID, bob.ID, link_info)

    assert not alice.qmemory.mem_positions[0].in_use
    assert not bob.qmemory.mem_positions[0].in_use

    ns.sim_reset()
    assert ns.sim_time() == 0
    netsquid_run(entdist.deliver(alice.ID, 0, bob.ID, 0, 0, 0))
    assert ns.sim_time() == 1000

    assert alice.qmemory.mem_positions[0].in_use
    assert bob.qmemory.mem_positions[0].in_use

    alice_qubit = alice.qmemory.peek([0])[0]
    bob_qubit = bob.qmemory.peek([0])[0]
    assert alice_qubit is not None
    assert bob_qubit is not None
    assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)


def test_deliver_depolar():
    alice, bob = create_n_nodes(2)

    entdist = create_entdist(nodes=[alice, bob])
    link_info = LhiLinkInfo.depolarise(
        cycle_time=10, prob_max_mixed=0.2, prob_success=1, state_delay=1000
    )
    entdist.add_sampler(alice.ID, bob.ID, link_info)

    assert not alice.qmemory.mem_positions[0].in_use
    assert not bob.qmemory.mem_positions[0].in_use

    ns.sim_reset()
    ns.set_qstate_formalism(QFormalism.DM)
    assert ns.sim_time() == 0
    netsquid_run(entdist.deliver(alice.ID, 0, bob.ID, 0, 0, 0))
    assert ns.sim_time() == 1000

    assert alice.qmemory.mem_positions[0].in_use
    assert bob.qmemory.mem_positions[0].in_use

    alice_qubit = alice.qmemory.peek([0])[0]
    bob_qubit = bob.qmemory.peek([0])[0]
    assert alice_qubit is not None
    assert bob_qubit is not None

    expected = 0.2 * TWO_MAX_MIXED + 0.8 * B00_DENS
    assert has_multi_state([alice_qubit, bob_qubit], expected)

    ns.set_qstate_formalism(QFormalism.KET)


def test_put_request():
    alice, bob = create_n_nodes(2)

    entdist = create_entdist(nodes=[alice, bob])

    assert len(entdist._requests) == 2
    assert len(entdist.get_requests(alice.ID)) == 0
    assert len(entdist.get_requests(bob.ID)) == 0

    request = create_request(alice.ID, bob.ID)
    entdist.put_request(request)

    assert len(entdist.get_requests(alice.ID)) == 1
    assert entdist.get_requests(alice.ID)[0] == request
    assert len(entdist.get_requests(bob.ID)) == 0


def test_put_request_many_nodes():
    n = 10
    nodes = create_n_nodes(n)

    entdist = create_entdist(nodes=nodes)

    assert len(entdist._requests) == n

    request01 = create_request(nodes[0].ID, nodes[1].ID)
    entdist.put_request(request01)

    request10 = create_request(nodes[1].ID, nodes[0].ID)
    entdist.put_request(request10)

    request05 = create_request(nodes[0].ID, nodes[5].ID)
    entdist.put_request(request05)

    assert entdist.get_requests(nodes[0].ID)[0] == request01
    assert entdist.get_requests(nodes[0].ID)[1] == request05
    assert entdist.get_requests(nodes[1].ID)[0] == request10

    request_invalid = create_request(nodes[0].ID, 100)
    with pytest.raises(ValueError):
        entdist.put_request(request_invalid)

    request_invalid = create_request(0, 0)
    with pytest.raises(ValueError):
        entdist.put_request(request_invalid)


def test_get_remote_request_for():
    alice, bob, charlie = create_n_nodes(3)
    entdist = create_entdist(nodes=[alice, bob, charlie])

    request_alice = create_request(alice.ID, bob.ID, 0, 1)
    entdist.put_request(request_alice)

    # Only Alice's request registered; no corresponding request from Bob yet.
    assert entdist.get_remote_request_for(request_alice) is None

    request_bob = create_request(bob.ID, alice.ID, 1, 0)
    entdist.put_request(request_bob)

    # Bob's first request corresponds to Alice's request.
    assert entdist.get_remote_request_for(request_alice) == 0

    invalid_request = create_request(100, 100)
    with pytest.raises(ValueError):
        entdist.get_remote_request_for(invalid_request)

    # Put 2 new identical requests for Alice.
    request_alice_1 = create_request(alice.ID, bob.ID, 2, 3)
    entdist.put_request(request_alice_1)
    request_alice_2 = create_request(alice.ID, bob.ID, 2, 3)
    entdist.put_request(request_alice_2)

    # Put 2 new requests for Bob. Only the 2nd corresponds to Alice.
    request_bob_1 = create_request(bob.ID, charlie.ID, 0, 0)
    entdist.put_request(request_bob_1)
    request_bob_2 = create_request(bob.ID, alice.ID, 3, 2)
    entdist.put_request(request_bob_2)

    # Bob's third (in total) request corresponds to Alice.
    assert entdist.get_remote_request_for(request_alice_1) == 2
    # Remove his first request.
    entdist.pop_request(bob.ID, 0)
    # The corresponding request is now 2nd in Bob's queue.
    assert entdist.get_remote_request_for(request_alice_1) == 1


def test_get_next_joint_request():
    alice, bob = create_n_nodes(2)
    entdist = create_entdist(nodes=[alice, bob])

    request_alice = create_request(alice.ID, bob.ID, 0, 0)
    entdist.put_request(request_alice)
    assert len(entdist.get_requests(alice.ID)) == 1
    assert len(entdist.get_requests(bob.ID)) == 0

    # Only Alice's request registered; no corresponding request from Bob yet.
    assert entdist.get_next_joint_request() is None

    # No requests should have been popped.
    assert len(entdist.get_requests(alice.ID)) == 1
    assert len(entdist.get_requests(bob.ID)) == 0

    request_bob = create_request(bob.ID, alice.ID, 0, 0)
    entdist.put_request(request_bob)
    assert len(entdist.get_requests(bob.ID)) == 1

    # Alice and Bob have corresponding requests.
    assert entdist.get_next_joint_request() is not None
    assert len(entdist.get_requests(alice.ID)) == 0
    assert len(entdist.get_requests(bob.ID)) == 0

    # Requests have been popped.
    assert entdist.get_next_joint_request() is None


def test_get_next_joint_request_2():
    alice, bob, charlie = create_n_nodes(3)
    entdist = create_entdist(nodes=[alice, bob, charlie])

    request_ab = create_request(alice.ID, bob.ID, 0, 0)
    request_ac = create_request(alice.ID, charlie.ID, 0, 1)
    entdist.put_request(request_ab)
    entdist.put_request(request_ac)

    assert entdist.get_next_joint_request() is None

    request_bc = create_request(bob.ID, charlie.ID, 0, 1)
    request_ba = create_request(bob.ID, alice.ID, 0, 0)
    entdist.put_request(request_bc)
    entdist.put_request(request_ba)

    # Alice and Bob have corresponding requests.
    assert entdist.get_next_joint_request() == create_joint_request(alice.ID, bob.ID)

    # Both Alice and Bob still have a request pending with Charlie.
    assert len(entdist.get_requests(alice.ID)) == 1
    assert len(entdist.get_requests(bob.ID)) == 1

    # No next joint request at this moment.
    assert entdist.get_next_joint_request() is None

    request_cb = create_request(charlie.ID, bob.ID, 1, 0)
    request_ca = create_request(charlie.ID, alice.ID, 1, 0)
    entdist.put_request(request_cb)
    entdist.put_request(request_ca)

    # Alice <-> Charlie
    # TODO: even though Bob put his request first, Alice's request is handled first
    # since Alice is first in the node list. Improve this!
    assert entdist.get_next_joint_request() == create_joint_request(
        alice.ID, charlie.ID, node1_pid=0, node2_pid=1
    )
    # Bob <-> Charlie
    assert entdist.get_next_joint_request() == create_joint_request(
        bob.ID, charlie.ID, node1_pid=0, node2_pid=1
    )


def test_serve_request():
    alice, bob = create_n_nodes(2, num_qubits=2)

    entdist = create_entdist(nodes=[alice, bob])
    link_info = LhiLinkInfo.perfect(1000)
    entdist.add_sampler(alice.ID, bob.ID, link_info)

    assert not alice.qmemory.mem_positions[0].in_use
    assert not bob.qmemory.mem_positions[0].in_use

    alice_mem = 0
    bob_mem = 0
    joint_request = create_joint_request(alice.ID, bob.ID, alice_mem, bob_mem)

    # Also create a joint request with invalid qubit location
    alice_mem_invalid = 3
    invalid_joint_request = create_joint_request(
        alice.ID, bob.ID, alice_mem_invalid, bob_mem
    )

    ns.sim_reset()
    assert ns.sim_time() == 0

    with pytest.raises(ValueError):
        netsquid_run(entdist.serve_request(invalid_joint_request))

    assert ns.sim_time() == 0

    netsquid_run(entdist.serve_request(joint_request))
    assert ns.sim_time() == 1000

    assert alice.qmemory.mem_positions[0].in_use
    assert bob.qmemory.mem_positions[0].in_use

    alice_qubit = alice.qmemory.peek([0])[0]
    bob_qubit = bob.qmemory.peek([0])[0]
    assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)

    alice_mem = 1
    bob_mem = 1
    joint_request = create_joint_request(alice.ID, bob.ID, alice_mem, bob_mem)
    netsquid_run(entdist.serve_request(joint_request))
    assert ns.sim_time() == 2000

    alice_qubit = alice.qmemory.peek([1])[0]
    bob_qubit = bob.qmemory.peek([1])[0]
    assert has_multi_state([alice_qubit, bob_qubit], B00_DENS)


def test_serve_request_multiple_nodes():
    alice, bob, charlie, david = create_n_nodes(4, num_qubits=2)

    entdist = create_entdist(nodes=[alice, bob, charlie, david])
    link_info = LhiLinkInfo.perfect(1000)
    for (node1, node2) in itertools.combinations([alice, bob, charlie, david], 2):
        entdist.add_sampler(node1.ID, node2.ID, link_info)

    req_ab = create_joint_request(alice.ID, bob.ID, 0, 0)
    req_ac = create_joint_request(alice.ID, charlie.ID, 1, 0)
    req_dc = create_joint_request(david.ID, charlie.ID, 0, 1)
    req_bd = create_joint_request(bob.ID, david.ID, 1, 1)

    ns.sim_reset()
    assert ns.sim_time() == 0
    netsquid_run(entdist.serve_request(req_ab))
    netsquid_run(entdist.serve_request(req_ac))
    netsquid_run(entdist.serve_request(req_dc))
    netsquid_run(entdist.serve_request(req_bd))
    assert ns.sim_time() == 4000

    alice_qubits = alice.qmemory.peek([0, 1])
    bob_qubits = bob.qmemory.peek([0, 1])
    charlie_qubits = charlie.qmemory.peek([0, 1])
    david_qubits = david.qmemory.peek([0, 1])
    assert has_multi_state([alice_qubits[0], bob_qubits[0]], B00_DENS)
    assert has_multi_state([alice_qubits[1], charlie_qubits[0]], B00_DENS)
    assert has_multi_state([david_qubits[0], charlie_qubits[1]], B00_DENS)
    assert has_multi_state([bob_qubits[1], david_qubits[1]], B00_DENS)


def test_entdist_run():
    alice, bob, charlie = create_n_nodes(3, num_qubits=2)
    entdist = create_entdist(nodes=[alice, bob, charlie])
    link_info = LhiLinkInfo.perfect(1000)
    for (node1, node2) in itertools.combinations([alice, bob, charlie], 2):
        entdist.add_sampler(node1.ID, node2.ID, link_info)

    req_ab = EntDistRequest(alice.ID, bob.ID, 0, [0], [0])
    req_ba = EntDistRequest(bob.ID, alice.ID, 0, [0], [0])

    req_ac = EntDistRequest(alice.ID, charlie.ID, 1, [1], [0])
    req_ca = EntDistRequest(charlie.ID, alice.ID, 0, [0], [1])

    req_bc = EntDistRequest(bob.ID, charlie.ID, 1, [1], [1])
    req_cb = EntDistRequest(charlie.ID, bob.ID, 1, [1], [1])

    ns.sim_reset()
    assert ns.sim_time() == 0
    assert not alice.qmemory.mem_positions[0].in_use
    assert not alice.qmemory.mem_positions[1].in_use
    assert not bob.qmemory.mem_positions[0].in_use
    assert not bob.qmemory.mem_positions[1].in_use
    assert not charlie.qmemory.mem_positions[0].in_use
    assert not charlie.qmemory.mem_positions[1].in_use

    # Bit of hack to send entrequest message to entdist
    alice_port = alice.add_ports("entdist_port")[0]
    bob_port = bob.add_ports("entdist_port")[0]
    charlie_port = charlie.add_ports("entdist_port")[0]

    alice_port.connect(entdist._comp.node_in_port(alice.name))
    bob_port.connect(entdist._comp.node_in_port(bob.name))
    charlie_port.connect(entdist._comp.node_in_port(charlie.name))

    alice_port.tx_output(Message(-1, -1, req_ab))
    bob_port.tx_output(Message(-1, -1, req_ba))
    alice_port.tx_output(Message(-1, -1, req_ac))
    charlie_port.tx_output(Message(-1, -1, req_ca))
    bob_port.tx_output(Message(-1, -1, req_bc))
    charlie_port.tx_output(Message(-1, -1, req_cb))

    entdist.start()
    ns.sim_run()

    # Since messages are sent instantly we can find the total time by 3 * 1000
    assert ns.sim_time() == 3000

    assert alice.qmemory.mem_positions[0].in_use
    assert alice.qmemory.mem_positions[1].in_use
    assert bob.qmemory.mem_positions[0].in_use
    assert bob.qmemory.mem_positions[1].in_use
    assert charlie.qmemory.mem_positions[0].in_use
    assert charlie.qmemory.mem_positions[1].in_use


if __name__ == "__main__":
    test_add_sampler()
    test_add_sampler_many_nodes()
    test_sample_perfect()
    test_sample_depolar()
    test_create_epr_pair_with_state()
    test_deliver_perfect()
    test_deliver_depolar()
    test_put_request()
    test_put_request_many_nodes()
    test_get_remote_request_for()
    test_get_next_joint_request()
    test_get_next_joint_request_2()
    test_serve_request()
    test_serve_request_multiple_nodes()
    test_entdist_run()



# type: ignore
from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Type

import yaml
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.instructions import Instruction as NetSquidInstruction
from netsquid.components.models.qerrormodels import (
    DepolarNoiseModel,
    QuantumErrorModel,
    T1T2NoiseModel,
)
from netsquid_magic.state_delivery_sampler import (
    DepolariseWithFailureStateSamplerFactory,
    IStateDeliverySamplerFactory,
    PerfectStateSamplerFactory,
)
from pydantic import BaseModel as PydanticBaseModel

from qoala.lang.common import MultiQubit
from qoala.runtime.instructions import (
    INSTR_BICHROMATIC,
    INSTR_MEASURE_ALL,
    INSTR_ROT_X_ALL,
    INSTR_ROT_Y_ALL,
    INSTR_ROT_Z_ALL,
)
from qoala.runtime.lhi import (
    INSTR_MEASURE_INSTANT,
    LhiGateConfigInterface,
    LhiLatenciesConfigInterface,
    LhiLinkConfigInterface,
    LhiNetworkScheduleConfigInterface,
    LhiNetworkTimebin,
    LhiQubitConfigInterface,
    LhiTopologyConfigInterface,
)
from qoala.runtime.ntf import (
    GenericNtf,
    NtfInterface,
    NtfInterfaceConfigInterface,
    NvNtf,
    TrappedIonNtf,
)
from qoala.util.math import fidelity_to_prob_max_mixed

GENERIC_GATES = [
    "INSTR_INIT",
    "INSTR_ROT_X",
    "INSTR_ROT_Y",
    "INSTR_ROT_Z",
    "INSTR_X",
    "INSTR_Y",
    "INSTR_Z",
    "INSTR_H",
    "INSTR_MEASURE",
    "INSTR_MEASURE_INSTANT",
]

GENERIC_TWO_GATES = ["INSTR_CNOT", "INSTR_CZ"]

GENERIC_DEFAULT_T1 = int(1e9)
GENERIC_DEFAULT_T2 = int(1e8)

GENERIC_DEFAULT_ONE_GATE_DURATION = int(5e3)
GENERIC_DEFAULT_TWO_GATE_DURATION = int(200e3)

NV_COM_GATES = [
    "INSTR_INIT",
    "INSTR_ROT_X",
    "INSTR_ROT_Y",
    "INSTR_MEASURE",
    "INSTR_MEASURE_INSTANT",
]

NV_MEM_GATES = [
    "INSTR_INIT",
    "INSTR_ROT_X",
    "INSTR_ROT_Y",
    "INSTR_ROT_Z",
    "INSTR_MEASURE",
    "INSTR_MEASURE_INSTANT",
]

NV_TWO_GATES = ["INSTR_CXDIR", "INSTR_CYDIR"]

NV_DEFAULT_COM_GATE_DURATION = 300
NV_DEFAULT_MEM_GATE_DURATION = int(1.2e6)
NV_DEFAULT_TWO_GATE_DURATION = int(1e6)

NV_DEFAULT_COM_T1 = int(3.6e12)
NV_DEFAULT_COM_T2 = int(5e8)
NV_DEFAULT_MEM_T1 = int(3.5e13)
NV_DEFAULT_MEM_T2 = int(1e9)


TRI_SINGLE_GATES = [
    "INSTR_INIT",
    "INSTR_ROT_Z",
    "INSTR_MEASURE",
    "INSTR_MEASURE_INSTANT",
]

TRI_ALL_GATES = [
    "INSTR_INIT",
    "INSTR_MEASURE_ALL",
    "INSTR_ROT_X_ALL",
    "INSTR_ROT_Y_ALL",
    "INSTR_ROT_Z_ALL",
    "INSTR_BICHROMATIC",
]

TRI_DEFAULT_SINGLE_GATE_DURATION = 26_600
TRI_DEFAULT_ALL_GATE_DURATION = 107_000

TRI_DEFAULT_T1 = int(1e99)
TRI_DEFAULT_T2 = int(85e6)

# NV_LAB_COMM_T1 = 1e6
# NV_LAB_COMM_T2 = 1e6
# NV_LAB_MEM_T1 = 1e6
# NV_LAB_MEM_T2 = 1e6

# NV_LAB_COMM_INIT_DURATION = 1e3


class BaseModel(PydanticBaseModel):
    class Config:
        arbitrary_types_allowed = True


class InstrConfigRegistry(ABC):
    @classmethod
    @abstractmethod
    def map(cls) -> Dict[str, NetSquidInstruction]:
        raise NotImplementedError


class QubitConfigRegistry(ABC):
    @classmethod
    @abstractmethod
    def map(cls) -> Dict[str, LhiQubitConfigInterface]:
        raise NotImplementedError


class GateConfigRegistry(ABC):
    @classmethod
    @abstractmethod
    def map(cls) -> Dict[str, GateNoiseConfigInterface]:
        raise NotImplementedError


class NtfInterfaceRegistry(ABC):
    @classmethod
    @abstractmethod
    def map(cls) -> Dict[str, NtfInterface]:
        raise NotImplementedError


class SamplerFactoryRegistry(ABC):
    @classmethod
    @abstractmethod
    def map(cls) -> Dict[str, IStateDeliverySamplerFactory]:
        raise NotImplementedError


def _from_dict(dict: Any, typ: Any) -> Any:
    return typ(**dict)


def _from_file(path: str, typ: Any) -> Any:
    with open(path, "r") as f:
        raw_config = yaml.load(f, Loader=yaml.Loader)
        return _from_dict(raw_config, typ)


def _read_dict(path: str) -> Any:
    with open(path, "r") as f:
        return yaml.load(f, Loader=yaml.Loader)


class QubitNoiseConfigInterface:
    @classmethod
    @abstractmethod
    def from_dict(cls, dict: Any) -> QubitNoiseConfigInterface:
        raise NotImplementedError

    @abstractmethod
    def to_error_model(self) -> Type[QuantumErrorModel]:
        raise NotImplementedError

    @abstractmethod
    def to_error_model_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError


class QubitT1T2Config(QubitNoiseConfigInterface, BaseModel):
    T1: int
    T2: int

    @classmethod
    def from_file(cls, path: str) -> QubitT1T2Config:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def from_dict(cls, dict: Any) -> QubitT1T2Config:
        return QubitT1T2Config(**dict)

    def to_error_model(self) -> Type[QuantumErrorModel]:
        return T1T2NoiseModel

    def to_error_model_kwargs(self) -> Dict[str, Any]:
        return {"T1": self.T1, "T2": self.T2}


class QubitConfig(LhiQubitConfigInterface, BaseModel):
    is_communication: bool
    noise_config_cls: str
    noise_config: QubitNoiseConfigInterface

    @classmethod
    def from_file(
        cls, path: str, registry: Optional[List[Type[QubitConfigRegistry]]] = None
    ) -> QubitConfig:
        return cls.from_dict(_read_dict(path), registry)

    @classmethod
    def perfect_config(cls, is_communication: bool) -> QubitConfig:
        return QubitConfig(
            is_communication=is_communication,
            noise_config_cls="T1T2NoiseModel",
            noise_config=QubitT1T2Config(T1=0, T2=0),
        )

    @classmethod
    def t1t2_config(cls, is_communication: bool, T1: int, T2: int) -> QubitConfig:
        return QubitConfig(
            is_communication=is_communication,
            noise_config_cls="T1T2NoiseModel",
            noise_config=QubitT1T2Config(T1=T1, T2=T2),
        )

    @classmethod
    def from_dict(
        cls, dict: Any, registry: Optional[List[Type[QubitConfigRegistry]]] = None
    ) -> QubitConfig:
        is_communication = dict["is_communication"]
        raw_typ = dict["noise_config_cls"]

        # Try to get the type of the noise config class.
        typ: Optional[QubitNoiseConfigInterface] = None
        # First try custom registries.
        if registry is not None:
            try:
                for reg in registry:
                    if raw_typ in reg.map():
                        typ = reg.map()[raw_typ]
                        break
            except KeyError:
                pass
        # If not found in custom registries, try default registry.
        if typ is None:
            try:
                typ = DefaultQubitConfigRegistry.map()[raw_typ]
            except KeyError:
                raise RuntimeError("invalid qubit noise class type")

        raw_noise_config = dict["noise_config"]
        noise_config = typ.from_dict(raw_noise_config)
        return QubitConfig(
            is_communication=is_communication,
            noise_config_cls=raw_typ,
            noise_config=noise_config,
        )

    def to_is_communication(self) -> bool:
        return self.is_communication

    def to_error_model(self) -> Type[QuantumErrorModel]:
        return self.noise_config.to_error_model()

    def to_error_model_kwargs(self) -> Dict[str, Any]:
        return self.noise_config.to_error_model_kwargs()


class GateNoiseConfigInterface:
    @classmethod
    @abstractmethod
    def from_dict(cls, dict: Any) -> GateNoiseConfigInterface:
        raise NotImplementedError

    @abstractmethod
    def to_duration(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def to_error_model(self) -> Type[QuantumErrorModel]:
        raise NotImplementedError

    @abstractmethod
    def to_error_model_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError


class GateDepolariseConfig(GateNoiseConfigInterface, BaseModel):
    duration: int
    depolarise_prob: float

    @classmethod
    def from_file(cls, path: str) -> GateDepolariseConfig:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def from_dict(cls, dict: Any) -> GateDepolariseConfig:
        return GateDepolariseConfig(**dict)

    def to_duration(self) -> int:
        return self.duration

    def to_error_model(self) -> Type[QuantumErrorModel]:
        return DepolarNoiseModel

    def to_error_model_kwargs(self) -> Dict[str, Any]:
        return {"depolar_rate": self.depolarise_prob, "time_independent": True}


class GateConfig(LhiGateConfigInterface, BaseModel):
    name: str
    noise_config_cls: str
    noise_config: GateNoiseConfigInterface

    @classmethod
    def from_file(
        cls, path: str, registry: Optional[List[Type[GateConfigRegistry]]] = None
    ) -> GateConfig:
        return cls.from_dict(_read_dict(path), registry)

    @classmethod
    def perfect_config(cls, name: str, duration: int) -> GateConfig:
        return GateConfig(
            name=name,
            noise_config_cls="GateDepolariseConfig",
            noise_config=GateDepolariseConfig(duration=duration, depolarise_prob=0),
        )

    @classmethod
    def with_depolar_prob(
        cls, name: str, duration: int, depolar_prob: float
    ) -> GateConfig:
        return GateConfig(
            name=name,
            noise_config_cls="GateDepolariseConfig",
            noise_config=GateDepolariseConfig(
                duration=duration, depolarise_prob=depolar_prob
            ),
        )

    @classmethod
    def with_single_gate_fidelity(
        cls, name: str, duration: int, fidelity: float
    ) -> GateConfig:
        depolar_prob = fidelity_to_prob_max_mixed(num_qubits=1, fid=fidelity)
        return cls.with_depolar_prob(name, duration, depolar_prob)

    @classmethod
    def with_two_gate_fidelity(
        cls, name: str, duration: int, fidelity: float
    ) -> GateConfig:
        depolar_prob = fidelity_to_prob_max_mixed(num_qubits=2, fid=fidelity)
        return cls.with_depolar_prob(name, duration, depolar_prob)

    @classmethod
    def with_all_qubit_gate_fidelity(
        cls, name: str, duration: int, fidelity: float, num_qubits: int
    ) -> GateConfig:
        depolar_prob = fidelity_to_prob_max_mixed(num_qubits=num_qubits, fid=fidelity)
        return cls.with_depolar_prob(name, duration, depolar_prob)

    @classmethod
    def from_dict(
        cls,
        dict: Any,
        registry: Optional[List[Type[GateConfigRegistry]]] = None,
    ) -> GateConfig:
        name = dict["name"]

        raw_typ = dict["noise_config_cls"]

        # Try to get the type of the noise config class.
        typ: Optional[GateNoiseConfigInterface] = None
        # First try custom registries.
        if registry is not None:
            try:
                for reg in registry:
                    if raw_typ in reg.map():
                        typ = reg.map()[raw_typ]
                        break
            except KeyError:
                pass
        # If not found in custom registries, try default registry.
        if typ is None:
            try:
                typ = DefaultGateConfigRegistry.map()[raw_typ]
            except KeyError:
                raise RuntimeError("invalid instruction type")

        raw_noise_config = dict["noise_config"]
        noise_config = typ.from_dict(raw_noise_config)
        return GateConfig(
            name=name,
            noise_config_cls=raw_typ,
            noise_config=noise_config,
        )

    def to_instruction(
        self,
        registry: Optional[List[Type[InstrConfigRegistry]]] = None,
    ) -> Type[NetSquidInstruction]:
        # Try to get the NetSquid Instruction class.
        instr: Optional[Type[NetSquidInstruction]] = None
        # First try custom registries.
        if registry is not None:
            try:
                for reg in registry:
                    if self.name in reg.map():
                        instr = reg.map()[self.name]
                        break
            except KeyError:
                pass
        # If not found in custom registries, try default registry.
        if instr is None:
            try:
                instr = DefaultInstrConfigRegistry.map()[self.name]
            except KeyError:
                raise RuntimeError("invalid instruction type")
        return instr

    def to_duration(self) -> int:
        return self.noise_config.to_duration()

    def to_error_model(self) -> Type[QuantumErrorModel]:
        return self.noise_config.to_error_model()

    def to_error_model_kwargs(self) -> Dict[str, Any]:
        return self.noise_config.to_error_model_kwargs()


class DefaultInstrConfigRegistry(InstrConfigRegistry):
    _MAP = {
        "INSTR_INIT": INSTR_INIT,
        "INSTR_X": INSTR_X,
        "INSTR_Y": INSTR_Y,
        "INSTR_Z": INSTR_Z,
        "INSTR_H": INSTR_H,
        "INSTR_ROT_X": INSTR_ROT_X,
        "INSTR_ROT_Y": INSTR_ROT_Y,
        "INSTR_ROT_Z": INSTR_ROT_Z,
        "INSTR_CNOT": INSTR_CNOT,
        "INSTR_CZ": INSTR_CZ,
        "INSTR_CXDIR": INSTR_CXDIR,
        "INSTR_CYDIR": INSTR_CYDIR,
        "INSTR_MEASURE": INSTR_MEASURE,
        "INSTR_MEASURE_INSTANT": INSTR_MEASURE_INSTANT,
        "INSTR_MEASURE_ALL": INSTR_MEASURE_ALL,
        "INSTR_ROT_X_ALL": INSTR_ROT_X_ALL,
        "INSTR_ROT_Y_ALL": INSTR_ROT_Y_ALL,
        "INSTR_ROT_Z_ALL": INSTR_ROT_Z_ALL,
        "INSTR_BICHROMATIC": INSTR_BICHROMATIC,
    }

    @classmethod
    def map(cls) -> Dict[str, NetSquidInstruction]:
        return cls._MAP


class DefaultQubitConfigRegistry(QubitConfigRegistry):
    _MAP = {
        "QubitT1T2Config": QubitT1T2Config,
    }

    @classmethod
    def map(cls) -> Dict[str, QubitNoiseConfigInterface]:
        return cls._MAP


class DefaultGateConfigRegistry(GateConfigRegistry):
    _MAP = {
        "GateDepolariseConfig": GateDepolariseConfig,
    }

    @classmethod
    def map(cls) -> Dict[str, GateNoiseConfigInterface]:
        return cls._MAP


class DefaultNtfRegistry(NtfInterfaceRegistry):
    _MAP = {
        "GenericNtf": GenericNtf,
        "NvNtf": NvNtf,
        "TrappedIonNtf": TrappedIonNtf,
    }

    @classmethod
    def map(cls) -> Dict[str, NtfInterface]:
        return cls._MAP


# Config classes directly used by Topology config.


class QubitIdConfig(BaseModel):
    qubit_id: int
    qubit_config: QubitConfig

    @classmethod
    def from_dict(cls, dict: Any) -> QubitIdConfig:
        return QubitIdConfig(
            qubit_id=dict["qubit_id"],
            qubit_config=QubitConfig.from_dict(dict["qubit_config"]),
        )


class SingleGateConfig(BaseModel):
    qubit_id: int
    gate_configs: List[GateConfig]

    @classmethod
    def from_dict(cls, dict: Any) -> SingleGateConfig:
        return SingleGateConfig(
            qubit_id=dict["qubit_id"],
            gate_configs=[GateConfig.from_dict(d) for d in dict["gate_configs"]],
        )


class MultiGateConfig(BaseModel):
    qubit_ids: List[int]
    gate_configs: List[GateConfig]

    @classmethod
    def from_dict(cls, dict: Any) -> MultiGateConfig:
        return MultiGateConfig(
            qubit_ids=dict["qubit_ids"],
            gate_configs=[GateConfig.from_dict(d) for d in dict["gate_configs"]],
        )


class AllQubitGateConfig(BaseModel):
    gate_configs: List[GateConfig]

    @classmethod
    def from_dict(cls, dict: Any) -> AllQubitGateConfig:
        return AllQubitGateConfig(
            gate_configs=[GateConfig.from_dict(d) for d in dict["gate_configs"]],
        )


# Topology config.


class TopologyConfig(BaseModel, LhiTopologyConfigInterface):
    qubits: List[QubitIdConfig]
    single_gates: List[SingleGateConfig]
    multi_gates: List[MultiGateConfig]
    all_qubit_gates: Optional[AllQubitGateConfig] = None

    @classmethod
    def from_file(cls, path: str) -> TopologyConfig:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def uniform_t1t2_qubits_perfect_gates(
        cls,
        num_qubits: int,
        t1: int,
        t2: int,
        single_instructions: List[str],
        single_duration: int,
        two_instructions: List[str],
        two_duration: int,
        all_qubit_gate_instructions: List[str] = None,
        all_qubit_gate_duration: int = 0,
    ) -> TopologyConfig:
        qubits = [
            QubitIdConfig(
                qubit_id=i,
                qubit_config=QubitConfig.t1t2_config(
                    is_communication=True, T1=t1, T2=t2
                ),
            )
            for i in range(num_qubits)
        ]

        single_gates = [
            SingleGateConfig(
                qubit_id=i,
                gate_configs=[
                    GateConfig.perfect_config(name=name, duration=single_duration)
                    for name in single_instructions
                ],
            )
            for i in range(num_qubits)
        ]

        multi_gates = []
        for i in range(num_qubits):
            for j in range(num_qubits):
                if i == j:
                    continue
                cfg = MultiGateConfig(
                    qubit_ids=[i, j],
                    gate_configs=[
                        GateConfig.perfect_config(name=name, duration=two_duration)
                        for name in two_instructions
                    ],
                )
                multi_gates.append(cfg)

        all_qubit_gates = None
        if all_qubit_gate_instructions is not None:
            all_qubit_gates = AllQubitGateConfig(
                gate_configs=[
                    GateConfig.perfect_config(
                        name=name, duration=all_qubit_gate_duration
                    )
                    for name in all_qubit_gate_instructions
                ]
            )

        return TopologyConfig(
            qubits=qubits,
            single_gates=single_gates,
            multi_gates=multi_gates,
            all_qubit_gates=all_qubit_gates,
        )

    @classmethod
    def perfect_config_uniform(
        cls,
        num_qubits: int,
        single_instructions: List[str],
        single_duration: int,
        two_instructions: List[str],
        two_duration: int,
        all_qubit_gate_instructions: List[str] = None,
        all_qubit_gate_duration: int = 0,
    ) -> TopologyConfig:
        return cls.uniform_t1t2_qubits_perfect_gates(
            num_qubits=num_qubits,
            t1=0,
            t2=0,
            single_instructions=single_instructions,
            single_duration=single_duration,
            two_instructions=two_instructions,
            two_duration=two_duration,
            all_qubit_gate_instructions=all_qubit_gate_instructions,
            all_qubit_gate_duration=all_qubit_gate_duration,
        )

    @classmethod
    def perfect_config_uniform_default_params(cls, num_qubits: int) -> TopologyConfig:
        return cls.perfect_config_uniform(
            num_qubits=num_qubits,
            single_instructions=GENERIC_GATES,
            single_duration=GENERIC_DEFAULT_ONE_GATE_DURATION,
            two_instructions=GENERIC_TWO_GATES,
            two_duration=GENERIC_DEFAULT_TWO_GATE_DURATION,
        )

    @classmethod
    def uniform_t1t2_qubits_perfect_gates_default_params(
        cls, num_qubits: int, t1: int, t2: int
    ) -> TopologyConfig:
        return cls.uniform_t1t2_qubits_perfect_gates(
            num_qubits=num_qubits,
            t1=t1,
            t2=t2,
            single_instructions=[
                "INSTR_INIT",
                "INSTR_ROT_X",
                "INSTR_ROT_Y",
                "INSTR_ROT_Z",
                "INSTR_X",
                "INSTR_Y",
                "INSTR_Z",
                "INSTR_H",
                "INSTR_MEASURE",
                "INSTR_MEASURE_INSTANT",
            ],
            single_duration=GENERIC_DEFAULT_ONE_GATE_DURATION,
            two_instructions=["INSTR_CNOT", "INSTR_CZ"],
            two_duration=GENERIC_DEFAULT_TWO_GATE_DURATION,
        )

    @classmethod
    def config_star(
        cls,
        num_qubits: int,
        comm_t1: int,
        comm_t2: int,
        mem_t1: int,
        mem_t2: int,
        comm_init_fidelity: float,
        comm_meas_fidelity: float,
        comm_gate_fidelity: float,
        comm_instructions: List[str],
        comm_duration: int,
        mem_instructions: List[str],
        mem_duration: int,
        two_instructions: List[str],
        two_duration: int,
    ) -> TopologyConfig:
        pass

    @classmethod
    def perfect_config_star(
        cls,
        num_qubits: int,
        comm_instructions: List[str],
        comm_duration: int,
        mem_instructions: List[str],
        mem_duration: int,
        two_instructions: List[str],
        two_duration: int,
    ) -> TopologyConfig:
        # comm qubit
        qubits = [
            QubitIdConfig(
                qubit_id=0,
                qubit_config=QubitConfig.perfect_config(is_communication=True),
            )
        ]
        # mem qubits
        qubits += [
            QubitIdConfig(
                qubit_id=i,
                qubit_config=QubitConfig.perfect_config(is_communication=False),
            )
            for i in range(1, num_qubits)
        ]

        # comm gate
        single_gates = [
            SingleGateConfig(
                qubit_id=0,
                gate_configs=[
                    GateConfig.perfect_config(name=name, duration=comm_duration)
                    for name in comm_instructions
                ],
            )
        ]
        # mem gates
        single_gates += [
            SingleGateConfig(
                qubit_id=i,
                gate_configs=[
                    GateConfig.perfect_config(name=name, duration=mem_duration)
                    for name in mem_instructions
                ],
            )
            for i in range(1, num_qubits)
        ]

        multi_gates = [
            MultiGateConfig(
                qubit_ids=[0, i],
                gate_configs=[
                    GateConfig.perfect_config(name=name, duration=two_duration)
                    for name in two_instructions
                ],
            )
            for i in range(1, num_qubits)
        ]

        return TopologyConfig(
            qubits=qubits, single_gates=single_gates, multi_gates=multi_gates
        )

    @classmethod
    def perfect_nv_default_params(cls, num_qubits: int) -> TopologyConfig:
        return cls.perfect_config_star(
            num_qubits=num_qubits,
            comm_instructions=NV_COM_GATES,
            comm_duration=NV_DEFAULT_COM_GATE_DURATION,
            mem_instructions=NV_MEM_GATES,
            mem_duration=NV_DEFAULT_MEM_GATE_DURATION,
            two_instructions=NV_TWO_GATES,
            two_duration=NV_DEFAULT_TWO_GATE_DURATION,
        )

    @classmethod
    def from_nv_params(cls, num_qubits: int, params: NvParams) -> TopologyConfig:
        return (
            TopologyConfigBuilder()
            .star_topology()
            .num_qubits(num_qubits)
            .default_nv_gates()
            .comm_t1(params.comm_t1)
            .comm_t2(params.comm_t2)
            .mem_t1(params.mem_t1)
            .mem_t2(params.mem_t2)
            .all_comm_gates_fidelity(params.comm_gate_fidelity)
            .all_comm_gates_duration(params.comm_gate_duration)
            .comm_gate_fidelity("INSTR_INIT", params.comm_init_fidelity)
            .comm_gate_duration("INSTR_INIT", params.comm_init_duration)
            .comm_gate_fidelity("INSTR_MEASURE", params.comm_meas_fidelity)
            .comm_gate_duration("INSTR_MEASURE", params.comm_meas_duration)
            .comm_gate_fidelity("INSTR_MEASURE_INSTANT", params.comm_meas_fidelity)
            .comm_gate_duration("INSTR_MEASURE_INSTANT", params.comm_meas_duration)
            .all_mem_gates_fidelity(params.mem_gate_fidelity)
            .all_mem_gates_duration(params.mem_gate_duration)
            .mem_gate_fidelity("INSTR_INIT", params.mem_init_fidelity)
            .mem_gate_duration("INSTR_INIT", params.mem_init_duration)
            .mem_gate_fidelity("INSTR_MEASURE", params.mem_meas_fidelity)
            .mem_gate_duration("INSTR_MEASURE", params.mem_meas_duration)
            .mem_gate_fidelity("INSTR_MEASURE_INSTANT", params.mem_meas_fidelity)
            .mem_gate_duration("INSTR_MEASURE_INSTANT", params.mem_meas_duration)
            .all_two_gates_fidelity(params.two_gate_fidelity)
            .all_two_gates_duration(params.two_gate_duration)
            .build()
        )

    @classmethod
    def perfect_tri_default_params(cls, num_qubits: int) -> TopologyConfig:
        return cls.perfect_config_uniform(
            num_qubits=num_qubits,
            single_instructions=TRI_SINGLE_GATES,
            single_duration=TRI_DEFAULT_SINGLE_GATE_DURATION,
            two_instructions=[],
            two_duration=0,
            all_qubit_gate_instructions=TRI_ALL_GATES,
            all_qubit_gate_duration=TRI_DEFAULT_ALL_GATE_DURATION,
        )

    # @classmethod
    # def nv_lab_setup(cls, num_qubits: int) -> TopologyConfig:
    #     # comm qubit
    #     qubits = [
    #         QubitIdConfig(
    #             qubit_id=0,
    #             qubit_config=QubitConfig.t1t2_config(
    #                 is_communication=True, T1=NV_LAB_COMM_T1, T2=NV_LAB_COMM_T2
    #             ),
    #         )
    #     ]
    #     # mem qubits
    #     qubits += [
    #         QubitIdConfig(
    #             qubit_id=i,
    #             qubit_config=QubitConfig.t1t2_config(
    #                 is_communication=False, T1=NV_LAB_MEM_T1, T2=NV_LAB_MEM_T2
    #             ),
    #         )
    #         for i in range(1, num_qubits)
    #     ]
    #     comm_gates: List[SingleGateConfig] = []
    #     comm_gates += [
    #         SingleGateConfig(
    #             qubit_id=0,
    #             gate_configs=[
    #                 GateConfig.perfect_config(
    #                     name=name, duration=NV_LAB_COMM_INIT_DURATION
    #                 )
    #                 for name in comm_instructions
    #             ],
    #         )
    #     ]

    @classmethod
    def from_dict(cls, dict: Any) -> TopologyConfig:
        raw_qubits = dict["qubits"]
        qubits = [QubitIdConfig.from_dict(d) for d in raw_qubits]
        raw_single_gates = dict["single_gates"]
        single_gates = [SingleGateConfig.from_dict(d) for d in raw_single_gates]
        raw_multi_gates = dict["multi_gates"]
        multi_gates = [MultiGateConfig.from_dict(d) for d in raw_multi_gates]
        all_qubit_gates = None
        if "all_qubit_gates" in dict:
            raw_all_qubit_gates = dict["all_qubit_gates"]
            all_qubit_gates = AllQubitGateConfig.from_dict(raw_all_qubit_gates)
        return TopologyConfig(
            qubits=qubits,
            single_gates=single_gates,
            multi_gates=multi_gates,
            all_qubit_gates=all_qubit_gates,
        )

    def get_qubit_configs(self) -> Dict[int, LhiQubitConfigInterface]:
        infos: Dict[int, LhiQubitConfigInterface] = {}
        for cfg in self.qubits:
            infos[cfg.qubit_id] = cfg.qubit_config
        return infos

    def get_single_gate_configs(self) -> Dict[int, List[LhiGateConfigInterface]]:
        infos: Dict[int, List[LhiGateConfigInterface]] = {}
        for cfg in self.single_gates:
            infos[cfg.qubit_id] = cfg.gate_configs
        return infos

    def get_multi_gate_configs(
        self,
    ) -> Dict[MultiQubit, List[LhiGateConfigInterface]]:
        infos: Dict[MultiQubit, List[LhiGateConfigInterface]] = {}
        for cfg in self.multi_gates:
            infos[MultiQubit(cfg.qubit_ids)] = cfg.gate_configs
        return infos

    def get_all_qubit_gate_configs(self) -> List[LhiGateConfigInterface]:
        if self.all_qubit_gates is None:
            return []
        return self.all_qubit_gates.gate_configs


class TopologyConfigBuilder:
    def __init__(self) -> None:
        self._num_qubits: Optional[int] = None
        self._uniform: Optional[bool] = None
        self._comm_t1: Optional[int] = None
        self._comm_t2: Optional[int] = None
        self._mem_t1: Optional[int] = None
        self._mem_t2: Optional[int] = None
        self._comm_single_gates: Optional[List[str]] = None
        self._comm_single_gate_durations: Optional[Dict[str, int]] = None
        self._comm_single_gate_fidelities: Optional[Dict[str, float]] = None
        self._mem_single_gates: Optional[List[str]] = None
        self._mem_single_gate_durations: Optional[Dict[str, int]] = None
        self._mem_single_gate_fidelities: Optional[Dict[str, float]] = None
        self._two_gates: Optional[List[str]] = None
        self._two_gate_durations: Optional[Dict[str, int]] = None
        self._two_gate_fidelities: Optional[Dict[str, float]] = None

    def _build_qubits(self) -> List[QubitIdConfig]:
        if self._uniform:
            qubits = [
                QubitIdConfig(
                    qubit_id=i,
                    qubit_config=QubitConfig.t1t2_config(
                        is_communication=True, T1=self._comm_t1, T2=self._comm_t2
                    ),
                )
                for i in range(self._num_qubits)
            ]
        else:
            # comm qubit
            qubits = [
                QubitIdConfig(
                    qubit_id=0,
                    qubit_config=QubitConfig.t1t2_config(
                        is_communication=True, T1=self._comm_t1, T2=self._comm_t2
                    ),
                )
            ]
            # mem qubits
            qubits += [
                QubitIdConfig(
                    qubit_id=i,
                    qubit_config=QubitConfig.t1t2_config(
                        is_communication=False, T1=self._mem_t1, T2=self._mem_t2
                    ),
                )
                for i in range(1, self._num_qubits)
            ]
        return qubits

    def _build_single_gates(self) -> List[SingleGateConfig]:
        single_gates: List[SingleGateConfig] = []

        if self._uniform:
            for i in range(self._num_qubits):
                gate_cfgs: List[GateConfig] = []
                for gate in self._comm_single_gates:
                    dur = self._comm_single_gate_durations[gate]
                    fid = self._comm_single_gate_fidelities[gate]
                    gate_cfgs.append(
                        GateConfig.with_single_gate_fidelity(gate, dur, fid)
                    )
                single_gates.append(
                    SingleGateConfig(qubit_id=i, gate_configs=gate_cfgs)
                )
        else:
            gate_cfgs: List[GateConfig] = []
            for gate in self._comm_single_gates:
                dur = self._comm_single_gate_durations[gate]
                fid = self._comm_single_gate_fidelities[gate]
                gate_cfgs.append(GateConfig.with_single_gate_fidelity(gate, dur, fid))
            single_gates.append(SingleGateConfig(qubit_id=0, gate_configs=gate_cfgs))
            for i in range(1, self._num_qubits):
                gate_cfgs: List[GateConfig] = []
                for gate in self._mem_single_gates:
                    dur = self._mem_single_gate_durations[gate]
                    fid = self._mem_single_gate_fidelities[gate]
                    gate_cfgs.append(
                        GateConfig.with_single_gate_fidelity(gate, dur, fid)
                    )
                single_gates.append(
                    SingleGateConfig(qubit_id=i, gate_configs=gate_cfgs)
                )

        return single_gates

    def _build_multi_gates(self) -> List[MultiGateConfig]:
        multi_gates: List[MultiGateConfig] = []

        if self._uniform:
            for i in range(self._num_qubits):
                for j in range(self._num_qubits):
                    if i == j:
                        continue
                    gate_cfgs: List[GateConfig] = []
                    for gate in self._two_gates:
                        dur = self._two_gate_durations[gate]
                        fid = self._two_gate_fidelities[gate]
                        gate_cfgs.append(
                            GateConfig.with_two_gate_fidelity(gate, dur, fid)
                        )
                    cfg = MultiGateConfig(qubit_ids=[i, j], gate_configs=gate_cfgs)
                    multi_gates.append(cfg)
        else:
            for i in range(1, self._num_qubits):
                gate_cfgs: List[GateConfig] = []
                for gate in self._two_gates:
                    dur = self._two_gate_durations[gate]
                    fid = self._two_gate_fidelities[gate]
                    gate_cfgs.append(GateConfig.with_two_gate_fidelity(gate, dur, fid))
                cfg = MultiGateConfig(qubit_ids=[0, i], gate_configs=gate_cfgs)
                multi_gates.append(cfg)

        return multi_gates

    def build(self) -> TopologyConfig:
        assert self._num_qubits is not None
        assert self._uniform is not None
        assert self._comm_t1 is not None
        assert self._comm_t2 is not None
        assert self._mem_t1 is not None
        assert self._mem_t2 is not None
        assert self._comm_single_gates is not None
        assert self._comm_single_gate_durations is not None
        assert self._comm_single_gate_fidelities is not None
        assert self._mem_single_gates is not None
        assert self._mem_single_gate_durations is not None
        assert self._mem_single_gate_fidelities is not None
        assert self._two_gates is not None
        assert self._two_gate_durations is not None
        assert self._two_gate_fidelities is not None

        qubits = self._build_qubits()
        single_gates = self._build_single_gates()
        multi_gates = self._build_multi_gates()
        return TopologyConfig(
            qubits=qubits,
            single_gates=single_gates,
            multi_gates=multi_gates,
        )

    def num_qubits(self, num: int) -> TopologyConfigBuilder:
        self._num_qubits = num
        return self

    def uniform_topology(self) -> TopologyConfigBuilder:
        self._uniform = True
        return self

    def star_topology(self) -> TopologyConfigBuilder:
        self._uniform = False
        return self

    def qubit_t1(self, t1: int) -> TopologyConfigBuilder:
        self._comm_t1 = t1
        self._mem_t1 = t1
        return self

    def qubit_t2(self, t2: int) -> TopologyConfigBuilder:
        self._comm_t2 = t2
        self._mem_t2 = t2
        return self

    def comm_t1(self, t1: int) -> TopologyConfigBuilder:
        self._comm_t1 = t1
        return self

    def comm_t2(self, t2: int) -> TopologyConfigBuilder:
        self._comm_t2 = t2
        return self

    def mem_t1(self, t1: int) -> TopologyConfigBuilder:
        self._mem_t1 = t1
        return self

    def mem_t2(self, t2: int) -> TopologyConfigBuilder:
        self._mem_t2 = t2
        return self

    def no_decoherence(self) -> TopologyConfigBuilder:
        self._comm_t1 = 0
        self._comm_t2 = 0
        self._mem_t1 = 0
        self._mem_t2 = 0
        return self

    def default_generic_gates(self) -> TopologyConfigBuilder:
        self._comm_single_gates = GENERIC_GATES
        self._mem_single_gates = GENERIC_GATES
        self._two_gates = GENERIC_TWO_GATES
        return self

    def default_nv_gates(self) -> TopologyConfigBuilder:
        self._comm_single_gates = NV_COM_GATES
        self._mem_single_gates = NV_MEM_GATES
        self._two_gates = NV_TWO_GATES
        return self

    def zero_gate_durations(self) -> TopologyConfigBuilder:
        if self._comm_single_gate_durations is None:
            self._comm_single_gate_durations = {}
        if self._mem_single_gate_durations is None:
            self._mem_single_gate_durations = {}
        if self._two_gate_durations is None:
            self._two_gate_durations = {}

        for gate in self._comm_single_gates:
            self._comm_single_gate_durations[gate] = 0
        for gate in self._mem_single_gates:
            self._mem_single_gate_durations[gate] = 0
        for gate in self._two_gates:
            self._two_gate_durations[gate] = 0
        return self

    def perfect_gate_fidelities(self) -> TopologyConfigBuilder:
        if self._comm_single_gate_fidelities is None:
            self._comm_single_gate_fidelities = {}
        if self._mem_single_gate_fidelities is None:
            self._mem_single_gate_fidelities = {}
        if self._two_gate_fidelities is None:
            self._two_gate_fidelities = {}

        for gate in self._comm_single_gates:
            self._comm_single_gate_fidelities[gate] = 1
        for gate in self._mem_single_gates:
            self._mem_single_gate_fidelities[gate] = 1
        for gate in self._two_gates:
            self._two_gate_fidelities[gate] = 1
        return self

    def all_comm_gates_fidelity(self, fidelity: float) -> TopologyConfigBuilder:
        if self._comm_single_gate_fidelities is None:
            self._comm_single_gate_fidelities = {}
        for gate in self._comm_single_gates:
            self._comm_single_gate_fidelities[gate] = fidelity
        return self

    def comm_gate_fidelity(self, gate: str, fidelity: float) -> TopologyConfigBuilder:
        if self._comm_single_gate_fidelities is None:
            self._comm_single_gate_fidelities = {}
        self._comm_single_gate_fidelities[gate] = fidelity
        return self

    def all_comm_gates_duration(self, duration: int) -> TopologyConfigBuilder:
        if self._comm_single_gate_durations is None:
            self._comm_single_gate_durations = {}
        for gate in self._comm_single_gates:
            self._comm_single_gate_durations[gate] = duration
        return self

    def comm_gate_duration(self, gate: str, duration: int) -> TopologyConfigBuilder:
        if self._comm_single_gate_durations is None:
            self._comm_single_gate_durations = {}
        self._comm_single_gate_durations[gate] = duration
        return self

    def all_mem_gates_fidelity(self, fidelity: float) -> TopologyConfigBuilder:
        if self._mem_single_gate_fidelities is None:
            self._mem_single_gate_fidelities = {}
        for gate in self._mem_single_gates:
            self._mem_single_gate_fidelities[gate] = fidelity
        return self

    def mem_gate_fidelity(self, gate: str, fidelity: float) -> TopologyConfigBuilder:
        if self._mem_single_gate_fidelities is None:
            self._mem_single_gate_fidelities = {}
        self._mem_single_gate_fidelities[gate] = fidelity
        return self

    def all_mem_gates_duration(self, duration: int) -> TopologyConfigBuilder:
        if self._mem_single_gate_durations is None:
            self._mem_single_gate_durations = {}
        for gate in self._mem_single_gates:
            self._mem_single_gate_durations[gate] = duration
        return self

    def mem_gate_duration(self, gate: str, duration: int) -> TopologyConfigBuilder:
        if self._mem_single_gate_durations is None:
            self._mem_single_gate_durations = {}
        self._mem_single_gate_durations[gate] = duration
        return self

    def all_two_gates_fidelity(self, fidelity: float) -> TopologyConfigBuilder:
        if self._two_gate_fidelities is None:
            self._two_gate_fidelities = {}
        for gate in self._two_gates:
            self._two_gate_fidelities[gate] = fidelity
        return self

    def two_gate_fidelity(self, gate: str, fidelity: float) -> TopologyConfigBuilder:
        if self._two_gate_fidelities is None:
            self._two_gate_fidelities = {}
        self._two_gate_fidelities[gate] = fidelity
        return self

    def all_two_gates_duration(self, duration: int) -> TopologyConfigBuilder:
        if self._two_gate_durations is None:
            self._two_gate_durations = {}
        for gate in self._two_gates:
            self._two_gate_durations[gate] = duration
        return self

    def two_gate_duration(self, gate: str, duration: int) -> TopologyConfigBuilder:
        if self._two_gate_durations is None:
            self._two_gate_durations = {}
        self._two_gate_durations[gate] = duration
        return self


# TODO: actually use this class (currently ProcNodeConfig just uses a string for ntf)
class NtfConfig(NtfInterfaceConfigInterface, BaseModel):
    ntf_interface_cls: str
    ntf_interface: Type[NtfInterface]

    @classmethod
    def from_file(
        cls, path: str, registry: Optional[List[Type[NtfInterfaceRegistry]]] = None
    ) -> NtfConfig:
        return cls.from_dict(_read_dict(path), registry)

    @classmethod
    def from_cls_name(
        cls, name: str, registry: Optional[List[Type[NtfInterfaceRegistry]]] = None
    ) -> NtfConfig:
        # Try to get the type of the ntf interface class.
        typ: Optional[NtfInterface] = None
        if registry is not None:
            try:
                for reg in registry:
                    if name in reg.map():
                        typ = reg.map()[name]
                        break
            except KeyError:
                pass
        # If not found in custom registries, try default registry.
        if typ is None:
            try:
                typ = DefaultNtfRegistry.map()[name]
            except KeyError:
                raise RuntimeError("invalid NTF interface type")

        return NtfConfig(ntf_interface_cls=name, ntf_interface=typ)

    @classmethod
    def from_dict(
        cls, dict: Any, registry: Optional[List[Type[NtfInterfaceRegistry]]] = None
    ) -> NtfConfig:
        raw_typ = dict["ntf_interface_cls"]
        return cls.from_cls_name(raw_typ, registry)

    def to_ntf_interface(self) -> Type[NtfInterface]:
        return self.ntf_interface


@dataclass
class NvParams:
    # Communication qubit coherence
    comm_t1: float = 0.0
    comm_t2: float = 0.0

    # Communication qubit gate noise
    comm_init_fidelity: float = 1.0
    comm_meas_fidelity: float = 1.0
    comm_gate_fidelity: float = 1.0

    # Communication qubit gate duration
    comm_init_duration: float = 0.0
    comm_meas_duration: float = 0.0
    comm_gate_duration: float = 0.0

    # Memory qubit coherence
    mem_t1: float = 0.0
    mem_t2: float = 0.0

    # Memory qubit gate noise
    mem_init_fidelity: float = 1.0
    mem_meas_fidelity: float = 1.0
    mem_gate_fidelity: float = 1.0

    # Memory qubit gate duration
    mem_init_duration: float = 0.0
    mem_meas_duration: float = 0.0
    mem_gate_duration: float = 0.0

    # Two-qubit gate noise
    two_gate_fidelity: float = 1.0
    # Two qubit gate duration
    two_gate_duration: float = 0.0


class LatenciesConfig(BaseModel, LhiLatenciesConfigInterface):
    host_instr_time: float = 0.0  # duration of classical Host instr execution
    qnos_instr_time: float = 0.0  # duration of classical Qnos instr execution
    host_peer_latency: float = 0.0  # processing time for Host messages from remote node
    internal_sched_latency: float = 0  # processing time for messaging between node scheduler and processor schedulers

    @classmethod
    def from_file(cls, path: str) -> LatenciesConfig:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def from_dict(cls, dict: Any) -> LatenciesConfig:
        host_instr_time = 0.0
        if "host_instr_time" in dict:
            host_instr_time = dict["host_instr_time"]
        qnos_instr_time = 0.0
        if "qnos_instr_time" in dict:
            qnos_instr_time = dict["qnos_instr_time"]
        host_peer_latency = 0.0
        if "host_peer_latency" in dict:
            host_peer_latency = dict["host_peer_latency"]
        internal_sched_latency = 0.0
        if "internal_sched_latency" in dict:
            host_peer_latency = dict["internal_sched_latency"]
        return LatenciesConfig(
            host_instr_time=host_instr_time,
            qnos_instr_time=qnos_instr_time,
            host_peer_latency=host_peer_latency,
            internal_sched_latency=internal_sched_latency,
        )

    def get_host_instr_time(self) -> float:
        return self.host_instr_time

    def get_qnos_instr_time(self) -> float:
        return self.qnos_instr_time

    def get_host_peer_latency(self) -> float:
        return self.host_peer_latency

    def get_internal_sched_latency(self) -> float:
        return self.internal_sched_latency


class ProcNodeConfig(BaseModel):
    node_name: str
    node_id: int
    topology: TopologyConfig
    latencies: LatenciesConfig
    ntf: NtfConfig
    determ_sched: bool = True
    use_deadlines: bool = True
    fcfs: bool = False
    prio_epr: bool = False
    is_predictable: bool = False

    @classmethod
    def from_file(cls, path: str) -> ProcNodeConfig:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def from_dict(cls, dict: Any) -> ProcNodeConfig:
        node_name = dict["node_name"]
        node_id = dict["node_id"]
        topology = TopologyConfig.from_dict(dict["topology"])
        latencies = LatenciesConfig.from_dict(dict["latencies"])
        ntf = NtfConfig.from_dict(dict["ntf"])
        return ProcNodeConfig(
            node_name=node_name,
            node_id=node_id,
            topology=topology,
            latencies=latencies,
            ntf=ntf,
        )


class LinkSamplerConfigInterface:
    @classmethod
    @abstractmethod
    def from_dict(cls, dict: Any) -> LinkSamplerConfigInterface:
        raise NotImplementedError

    @abstractmethod
    def to_sampler_factory(self) -> Type[IStateDeliverySamplerFactory]:
        raise NotImplementedError

    @abstractmethod
    def to_sampler_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError


class PerfectSamplerConfig(LinkSamplerConfigInterface, BaseModel):
    cycle_time: float

    @classmethod
    def from_dict(cls, dict: Any) -> PerfectSamplerConfig:
        return PerfectSamplerConfig(**dict)

    def to_sampler_factory(self) -> Type[IStateDeliverySamplerFactory]:
        return PerfectStateSamplerFactory

    def to_sampler_kwargs(self) -> Dict[str, Any]:
        return {"cycle_time": self.cycle_time}


class DepolariseSamplerConfig(LinkSamplerConfigInterface, BaseModel):
    cycle_time: float
    prob_max_mixed: float
    prob_success: float

    @classmethod
    def from_dict(cls, dict: Any) -> DepolariseSamplerConfig:
        return DepolariseSamplerConfig(**dict)

    def to_sampler_factory(self) -> Type[IStateDeliverySamplerFactory]:
        return DepolariseWithFailureStateSamplerFactory

    def to_sampler_kwargs(self) -> Dict[str, Any]:
        return {
            "cycle_time": self.cycle_time,
            "prob_max_mixed": self.prob_max_mixed,
            "prob_success": self.prob_success,
        }


class DefaultSamplerConfigRegistry(SamplerFactoryRegistry):
    _MAP = {
        "PerfectSamplerConfig": PerfectSamplerConfig,
        "DepolariseSamplerConfig": DepolariseSamplerConfig,
    }

    @classmethod
    def map(cls) -> Dict[str, LinkSamplerConfigInterface]:
        return cls._MAP


class LinkConfig(LhiLinkConfigInterface, BaseModel):
    state_delay: float
    sampler_config_cls: str
    sampler_config: LinkSamplerConfigInterface

    @classmethod
    def from_file(cls, path: str) -> LinkConfig:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def perfect_config(cls, state_delay: float) -> LinkConfig:
        return LinkConfig(
            state_delay=state_delay,
            sampler_config_cls="PerfectSamplerConfig",
            sampler_config=PerfectSamplerConfig(cycle_time=0),
        )

    @classmethod
    def simple_depolarise_config(
        cls, fidelity: float, state_delay: float
    ) -> LinkConfig:
        prob_max_mixed = fidelity_to_prob_max_mixed(2, fidelity)
        sampler_config = DepolariseSamplerConfig(
            cycle_time=0, prob_max_mixed=prob_max_mixed, prob_success=1
        )
        return LinkConfig(
            state_delay=state_delay,
            sampler_config_cls="DepolariseSamplerConfig",
            sampler_config=sampler_config,
        )

    @classmethod
    def depolarise_config(
        cls,
        prob_max_mixed: float,
        attempt_success_prob: float,
        attempt_duration: float,
        state_delay: float,
    ) -> LinkConfig:
        sampler_config = DepolariseSamplerConfig(
            cycle_time=attempt_duration,
            prob_max_mixed=prob_max_mixed,
            prob_success=attempt_success_prob,
        )
        return LinkConfig(
            state_delay=state_delay,
            sampler_config_cls="DepolariseSamplerConfig",
            sampler_config=sampler_config,
        )

    @classmethod
    def from_dict(
        cls, dict: Any, registry: Optional[List[Type[SamplerFactoryRegistry]]] = None
    ) -> LinkConfig:
        state_delay = dict["state_delay"]
        raw_typ = dict["sampler_config_cls"]

        # Try to get the type of the sampler config class.
        typ: Optional[LinkSamplerConfigInterface] = None
        # First try custom registries.
        if registry is not None:
            try:
                for reg in registry:
                    if raw_typ in reg.map():
                        typ = reg.map()[raw_typ]
                        break
            except KeyError:
                pass
        # If not found in custom registries, try default registry.
        if typ is None:
            try:
                typ = DefaultSamplerConfigRegistry.map()[raw_typ]
            except KeyError:
                raise RuntimeError("invalid sampler config type")

        raw_sampler_config = dict["sampler_config"]
        sampler_config = typ.from_dict(raw_sampler_config)

        return LinkConfig(
            state_delay=state_delay,
            sampler_config_cls=raw_typ,
            sampler_config=sampler_config,
        )

    def to_sampler_factory(self) -> Type[IStateDeliverySamplerFactory]:
        return self.sampler_config.to_sampler_factory()

    def to_sampler_kwargs(self) -> Dict[str, Any]:
        return self.sampler_config.to_sampler_kwargs()

    def to_state_delay(self) -> float:
        return self.state_delay


class LinkBetweenNodesConfig(BaseModel):
    node_id1: int
    node_id2: int
    link_config: LinkConfig

    @classmethod
    def from_file(cls, path: str) -> LinkBetweenNodesConfig:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def from_dict(cls, dict: Any) -> LinkBetweenNodesConfig:
        return LinkBetweenNodesConfig(
            node_id1=dict["node_id1"],
            node_id2=dict["node_id2"],
            link_config=LinkConfig.from_dict(dict["link_config"]),
        )


class ClassicalConnectionConfig(BaseModel):
    node_id1: int
    node_id2: int
    latency: float

    @classmethod
    def from_nodes(
        cls, node_id1: int, node_id2: int, latency: float
    ) -> ClassicalConnectionConfig:
        return ClassicalConnectionConfig(
            node_id1=node_id1, node_id2=node_id2, latency=latency
        )

    @classmethod
    def from_file(cls, path: str) -> ClassicalConnectionConfig:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def from_dict(cls, dict: Any) -> ClassicalConnectionConfig:
        return ClassicalConnectionConfig(
            node_id1=dict["node_id1"],
            node_id2=dict["node_id2"],
            latency=dict["latency"],
        )


class NetworkScheduleConfig(BaseModel, LhiNetworkScheduleConfigInterface):
    bin_length: int
    first_bin: int
    bin_pattern: List[Tuple[int, int, int, int]]
    repeat_period: int

    @classmethod
    def from_file(cls, path: str) -> NetworkScheduleConfig:
        return cls.from_dict(_read_dict(path))

    @classmethod
    def from_dict(cls, dict: Any) -> NetworkScheduleConfig:
        return NetworkScheduleConfig(
            bin_length=dict["bin_length"],
            first_bin=dict["first_bin"],
            bin_pattern=dict["bin_pattern"],
            repeat_period=dict["repeat_period"],
        )

    def to_bin_length(self) -> int:
        return self.bin_length

    def to_first_bin(self) -> int:
        return self.first_bin

    def to_bin_pattern(self) -> List[LhiNetworkTimebin]:
        pattern = [
            LhiNetworkTimebin(frozenset({node1, node2}), {node1: pid1, node2: pid2})
            for (node1, pid1, node2, pid2) in self.bin_pattern
        ]
        return pattern

    def to_repeat_period(self) -> int:
        return self.repeat_period


class ProcNodeNetworkConfig(BaseModel):
    nodes: List[ProcNodeConfig]
    links: List[LinkBetweenNodesConfig]
    netschedule: Optional[NetworkScheduleConfig] = None
    cconns: List[ClassicalConnectionConfig] = None

    @classmethod
    def from_file(cls, path: str) -> ProcNodeNetworkConfig:
        return _from_file(path, ProcNodeNetworkConfig)  # type: ignore

    @classmethod
    def from_nodes_perfect_links(
        cls, nodes: List[ProcNodeConfig], link_duration: float
    ) -> ProcNodeNetworkConfig:
        links: List[LinkBetweenNodesConfig] = []
        for node1, node2 in itertools.combinations(nodes, 2):
            links.append(
                LinkBetweenNodesConfig(
                    node_id1=node1.node_id,
                    node_id2=node2.node_id,
                    link_config=LinkConfig.perfect_config(link_duration),
                )
            )
        return ProcNodeNetworkConfig(nodes=nodes, links=links)



from typing import Generator, List, Optional

import pytest
from netqasm.lang.subroutine import Subroutine

from pydynaa import EventExpression
from qoala.lang.ehi import UnitModule
from qoala.lang.hostlang import IqoalaVector
from qoala.lang.program import ProgramMeta, QoalaProgram
from qoala.lang.request import (
    CallbackType,
    EprRole,
    EprType,
    QoalaRequest,
    RequestRoutine,
    RequestVirtIdMapping,
)
from qoala.lang.routine import LocalRoutine, RoutineMetadata
from qoala.runtime.lhi import LhiTopology, LhiTopologyBuilder
from qoala.runtime.lhi_to_ehi import LhiConverter
from qoala.runtime.memory import ProgramMemory, RunningRequestRoutine
from qoala.runtime.message import Message, RrCallTuple
from qoala.runtime.ntf import GenericNtf
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.sharedmem import MemAddr
from qoala.sim.entdist.entdist import EntDistRequest
from qoala.sim.eprsocket import EprSocket
from qoala.sim.memmgr import AllocError, MemoryManager
from qoala.sim.netstack import NetstackInterface, NetstackLatencies, NetstackProcessor
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice, QDeviceCommand


class MockNetstackInterface(NetstackInterface):
    def __init__(
        self,
        qdevice: QDevice,
        memmgr: MemoryManager,
    ) -> None:
        self._qdevice = qdevice
        self._memmgr = memmgr

    def send_qnos_msg(self, msg: Message) -> None:
        pass

    @property
    def node_id(self) -> int:
        return 0


class MockQDevice(QDevice):
    def __init__(self, topology: LhiTopology) -> None:
        self._topology = topology

        self._executed_commands: List[QDeviceCommand] = []

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        pass

    def execute_commands(
        self, commands: List[QDeviceCommand]
    ) -> Generator[EventExpression, None, Optional[int]]:
        self._executed_commands.extend(commands)
        return None
        yield


def generic_topology(num_qubits: int) -> LhiTopology:
    # Instructions and durations are not needed for these tests.
    return LhiTopologyBuilder.perfect_uniform(
        num_qubits=num_qubits,
        single_instructions=[],
        single_duration=0,
        two_instructions=[],
        two_duration=0,
    )


def star_topology(num_qubits: int) -> LhiTopology:
    # Instructions and durations are not needed for these tests.
    return LhiTopologyBuilder.perfect_star(
        num_qubits=num_qubits,
        comm_instructions=[],
        comm_duration=0,
        mem_instructions=[],
        mem_duration=0,
        two_instructions=[],
        two_duration=0,
    )


def create_process(
    pid: int, unit_module: UnitModule, local_routines=None, request_routines=None
) -> QoalaProcess:
    program = QoalaProgram(
        blocks=[],
        local_routines=local_routines,
        meta=ProgramMeta.empty("prog"),
        request_routines=request_routines,
    )
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=ProgramInput({}),
        unit_module=unit_module,
    )
    mem = ProgramMemory(pid=pid)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def create_simple_request(
    remote_id: int,
    num_pairs: int,
    virt_ids: RequestVirtIdMapping,
    epr_socket_id: int = 0,
) -> QoalaRequest:
    return QoalaRequest(
        name="req",
        remote_id=remote_id,
        epr_socket_id=epr_socket_id,
        num_pairs=num_pairs,
        virt_ids=virt_ids,
        timeout=1000,
        fidelity=0.65,
        typ=EprType.CREATE_KEEP,
        role=EprRole.CREATE,
    )


def test__allocate_for_pair():
    topology = generic_topology(10)
    qdevice = MockQDevice(topology)
    ehi = LhiConverter.to_ehi(topology, ntf=GenericNtf())
    unit_module = UnitModule.from_full_ehi(ehi)
    memmgr = MemoryManager("alice", qdevice)

    process = create_process(0, unit_module)
    memmgr.add_process(process)

    interface = MockNetstackInterface(qdevice, memmgr)
    latencies = NetstackLatencies.all_zero()
    processor = NetstackProcessor(interface, latencies)

    request = create_simple_request(
        remote_id=1, num_pairs=5, virt_ids=RequestVirtIdMapping.from_str("all 0")
    )

    assert memmgr.phys_id_for(process.pid, 0) is None
    assert processor._allocate_for_pair(process, request, 0) == 0
    assert memmgr.phys_id_for(process.pid, 0) == 0

    assert memmgr.phys_id_for(process.pid, 1) is None
    with pytest.raises(AllocError):
        # Index 1 also requires virt ID but it's already alloacted
        processor._allocate_for_pair(process, request, 1)

    request2 = create_simple_request(
        remote_id=1, num_pairs=2, virt_ids=RequestVirtIdMapping.from_str("increment 1")
    )
    assert processor._allocate_for_pair(process, request2, index=0) == 1
    assert memmgr.phys_id_for(process.pid, virt_id=1) == 1
    assert processor._allocate_for_pair(process, request2, index=1) == 2
    assert memmgr.phys_id_for(process.pid, virt_id=2) == 2

    request3 = create_simple_request(
        remote_id=1,
        num_pairs=2,
        virt_ids=RequestVirtIdMapping.from_str("custom 5, 7"),
    )
    assert processor._allocate_for_pair(process, request3, index=0) == 5
    assert memmgr.phys_id_for(process.pid, virt_id=5) == 3
    assert processor._allocate_for_pair(process, request3, index=1) == 7
    assert memmgr.phys_id_for(process.pid, virt_id=7) == 4


def test__create_entdist_request():
    topology = generic_topology(5)
    qdevice = MockQDevice(topology)
    ehi = LhiConverter.to_ehi(topology, ntf=GenericNtf())
    unit_module = UnitModule.from_full_ehi(ehi)
    memmgr = MemoryManager("alice", qdevice)

    process = create_process(0, unit_module)
    memmgr.add_process(process)

    interface = MockNetstackInterface(qdevice, memmgr)
    latencies = NetstackLatencies.all_zero()
    processor = NetstackProcessor(interface, latencies)

    request = create_simple_request(
        remote_id=1,
        num_pairs=5,
        virt_ids=RequestVirtIdMapping.from_str("all 0"),
        epr_socket_id=7,
    )

    phys_id = memmgr.allocate(process.pid, 3)
    process.epr_sockets[7] = EprSocket(7, 1, 0, 42, 1.0)
    assert processor._create_entdist_request(process, request, 3) == EntDistRequest(
        local_node_id=interface.node_id,
        remote_node_id=1,
        local_qubit_id=phys_id,
        local_pid=0,
        remote_pid=42,
    )


def test_instantiate():
    topology = generic_topology(5)
    qdevice = MockQDevice(topology)
    ehi = LhiConverter.to_ehi(topology, ntf=GenericNtf())
    unit_module = UnitModule.from_full_ehi(ehi)
    memmgr = MemoryManager("alice", qdevice)

    subrt = Subroutine()
    metadata = RoutineMetadata.use_none()
    local_routine = LocalRoutine(
        "subrt1", subrt, return_vars=[IqoalaVector("res", 3)], metadata=metadata
    )

    request = QoalaRequest(
        name="req",
        remote_id=1,
        epr_socket_id=0,
        num_pairs=3,
        virt_ids=RequestVirtIdMapping.from_str("all 0"),
        timeout=1000,
        fidelity=0.65,
        typ=EprType.CREATE_KEEP,
        role=EprRole.CREATE,
    )
    routine = RequestRoutine(
        "req",
        request,
        [IqoalaVector("outcomes", 1)],
        CallbackType.WAIT_ALL,
        "subrt1",
    )

    process = create_process(
        0,
        unit_module,
        local_routines={"subrt1": local_routine},
        request_routines={"req": routine},
    )
    memmgr.add_process(process)

    interface = MockNetstackInterface(qdevice, memmgr)
    latencies = NetstackLatencies.all_zero()
    processor = NetstackProcessor(interface, latencies)

    rrcall = RrCallTuple(
        "req",
        MemAddr(0),
        MemAddr(1),
        [MemAddr(2), MemAddr(3), MemAddr(4)],
        [MemAddr(5), MemAddr(6), MemAddr(7)],
    )

    processor.instantiate_routine(process, rrcall, args={})

    runningReq = process.qnos_mem.get_running_request_routine("req")
    assert runningReq == RunningRequestRoutine(
        routine=routine,
        params_addr=rrcall.input_addr,
        result_addr=rrcall.result_addr,
        cb_input_addrs=rrcall.cb_input_addrs,
        cb_output_addrs=rrcall.cb_output_addrs,
    )

    assert runningReq.routine is not routine


if __name__ == "__main__":
    test__allocate_for_pair()
    test__create_entdist_request()
    test_instantiate()



from __future__ import annotations

import json
import math
import os
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from datetime import datetime
from pathlib import Path
from typing import List

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.logging import LogManager
from qoala.util.runner import (
    SchedulerType,
    run_single_node_app,
    run_two_node_app_separate_inputs,
    run_two_node_app_separate_inputs_plus_constant_tasks,
)


def create_procnode_cfg(
    name: str, id: int, t1: float, t2: float, determ: bool, deadlines: bool, fcfs: bool
) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.uniform_t1t2_qubits_perfect_gates_default_params(
            5, t1, t2
        ),
        latencies=LatenciesConfig(qnos_instr_time=1000, host_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
        determ_sched=determ,
        use_deadlines=deadlines,
        fcfs=fcfs,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class ProgResult:
    alice_results: BatchResult
    bob_results: BatchResult
    qpu_waits: List[float]
    total_duration: float


def run_apps(
    state: int,
    t1: float,
    t2: float,
    cc_latency: float,
    num_const_tasks: int,
    busy_duration: float,
    const_period: float,
    const_start: float,
    sched_typ: SchedulerType,
) -> ProgResult:
    ns.sim_reset()

    alice_id = 1
    bob_id = 0

    fcfs = sched_typ == SchedulerType.FCFS

    # Alice never uses FCFS!
    alice_node_cfg = create_procnode_cfg(
        "alice", alice_id, t1, t2, determ=True, deadlines=True, fcfs=False
    )
    bob_node_cfg = create_procnode_cfg(
        "bob", bob_id, t1, t2, determ=True, deadlines=True, fcfs=fcfs
    )

    cconn = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, cc_latency)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )
    network_cfg.cconns = [cconn]

    alice_program = load_program("programs/controller.iqoala")
    bob_program = load_program("programs/interactive_quantum.iqoala")
    busy_program = load_program("programs/cpu_busy.iqoala")

    alice_inputs = [ProgramInput({"bob_id": bob_id})]

    def measure_state(prepare_state: int) -> int:
        return {0: 1, 1: 0, 2: 3, 3: 2, 4: 5, 5: 4}[prepare_state]

    bob_inputs = [
        ProgramInput(
            {
                "alice_id": alice_id,
                "prepare_state": state,
                "measure_state": measure_state(state),
            },
        )
    ]

    busy_inputs = [
        ProgramInput({"duration": busy_duration}) for _ in range(num_const_tasks)
    ]

    if sched_typ == SchedulerType.QOALA or sched_typ == SchedulerType.FCFS:
        app_result = run_two_node_app_separate_inputs_plus_constant_tasks(
            num_iterations=1,
            num_const_tasks=num_const_tasks,
            node1="alice",
            node2="bob",
            prog_node1=alice_program,
            prog_node1_inputs=alice_inputs,
            prog_node2=bob_program,
            prog_node2_inputs=bob_inputs,
            const_prog_node2=busy_program,
            const_prog_node2_inputs=busy_inputs,
            const_period=int(const_period),
            const_start=int(const_start),
            network_cfg=network_cfg,
            sched_typ=sched_typ,
            linear=True,
        )
        alice_result = app_result.batch_results["alice"]
        bob_result = app_result.batch_results["bob"]

        return ProgResult(alice_result, bob_result, [], app_result.total_duration)
    else:
        app_result_quantum = run_two_node_app_separate_inputs(
            num_iterations=1,
            programs={"alice": alice_program, "bob": bob_program},
            program_inputs={"alice": alice_inputs, "bob": bob_inputs},
            network_cfg=network_cfg,
            linear=True,
        )
        app_result_classical = run_single_node_app(
            num_iterations=num_const_tasks,
            program_name="bob",
            program=busy_program,
            program_input=ProgramInput({"duration": busy_duration}),
            network_cfg=network_cfg,
            linear=True,
        )
        alice_result = app_result_quantum.batch_results["alice"]
        bob_result = app_result_quantum.batch_results["bob"]

        total_duration = (
            app_result_classical.total_duration + app_result_quantum.total_duration
        )

        return ProgResult(alice_result, bob_result, [], total_duration)


@dataclass
class DataPoint:
    sched_typ: str
    busy_factor: float
    busy_duration: float
    succ_prob: float
    succ_prob_lower: float
    succ_prob_upper: float
    makespan: float


@dataclass
class DataMeta:
    timestamp: str
    sim_duration: float
    num_iterations: int
    t1: float
    t2: float
    latency_factor: float
    num_const_tasks: int
    const_rate_factor: float
    cc_latency: float
    const_period: float
    const_start: float
    busy_factors: List[float]


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def wilson_score_interval(p_hat, n, z):
    denominator = 1 + z**2 / n
    centre_adjusted_probability = p_hat + z**2 / (2 * n)
    adjusted_standard_deviation = z * math.sqrt(
        (p_hat * (1 - p_hat) + z**2 / (4 * n)) / n
    )

    lower_bound = (
        centre_adjusted_probability - adjusted_standard_deviation
    ) / denominator
    upper_bound = (
        centre_adjusted_probability + adjusted_standard_deviation
    ) / denominator

    return (lower_bound, upper_bound)


def get_datapoint(
    num_iterations: int,
    t1: float,
    t2: float,
    cc_latency: float,
    num_const_tasks: int,
    const_period: float,
    const_start: float,
    sched_typ: SchedulerType,
    busy_factor: float,
) -> DataPoint:
    successes: List[bool] = []
    makespans: List[float] = []

    busy_duration = busy_factor * cc_latency

    for i in range(num_iterations):
        result = run_apps(
            state=i % 6,
            t1=t1,
            t2=t2,
            cc_latency=cc_latency,
            num_const_tasks=num_const_tasks,
            busy_duration=busy_duration,
            const_period=const_period,
            const_start=const_start,
            sched_typ=sched_typ,
        )
        program_results = result.bob_results.results
        outcomes = [result.values["outcome"] for result in program_results]
        assert len(outcomes) == 1
        successes.append(outcomes[0] == 1)
        makespans.append(result.total_duration)

    avg_succ_prob = sum([s for s in successes if s]) / len(successes)
    succ_prob_lower, succ_prob_upper = wilson_score_interval(
        p_hat=avg_succ_prob, n=len(successes), z=1.96
    )

    total_makespan = sum(makespans)

    return DataPoint(
        sched_typ=sched_typ.name,
        busy_factor=busy_factor,
        busy_duration=busy_duration,
        succ_prob=avg_succ_prob,
        succ_prob_lower=succ_prob_lower,
        succ_prob_upper=succ_prob_upper,
        makespan=total_makespan,
    )


def run(
    output_dir: str,
    sched_typ: SchedulerType,
    num_iterations: int,
    dump: bool = True,
    log: bool = False,
):
    if log:
        LogManager.set_log_level("DEBUG")
        LogManager.log_to_file("classical_multitasking_2.log")
        LogManager.set_task_log_level("DEBUG")
        LogManager.log_tasks_to_file("classical_multitasking_tasks_2.log")

    start_time = time.time()
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    # print(timestamp)

    data_points: List[DataPoint] = []

    # Constants
    t1 = 1e10
    t2 = 1e8
    latency_factor = 0.1
    num_const_tasks = 100
    const_rate_factor = 10

    cc_latency = latency_factor * t2
    const_period = cc_latency / const_rate_factor
    const_start = cc_latency

    # Variables
    busy_factors = [
        # 0.05,
        0.1,
        0.15,
        0.2,
        0.3,
        0.5,
        1,
        2,
        5,
        10,
    ]  # fraction of cc_latency
    # busy_factors = [0.5]
    # busy_factors = [0.1]

    for i, busy_factor in enumerate(busy_factors):
        print(
            f"{sched_typ:30}: data point {i}/{len(busy_factors)}", end="\r", flush=True
        )
        data_point = get_datapoint(
            num_iterations=num_iterations,
            t1=t1,
            t2=t2,
            cc_latency=cc_latency,
            num_const_tasks=num_const_tasks,
            const_period=const_period,
            const_start=const_start,
            sched_typ=sched_typ,
            busy_factor=busy_factor,
        )

        data_points.append(data_point)

    end_time = time.time()
    sim_duration = end_time - start_time

    meta = DataMeta(
        timestamp=timestamp,
        sim_duration=sim_duration,
        num_iterations=num_iterations,
        t1=t1,
        t2=t2,
        latency_factor=latency_factor,
        num_const_tasks=num_const_tasks,
        const_rate_factor=const_rate_factor,
        cc_latency=cc_latency,
        const_period=const_period,
        const_start=const_start,
        busy_factors=busy_factors,
    )

    if dump:
        data = Data(meta=meta, data_points=data_points)

        json_data = asdict(data)

        abs_dir = relative_to_cwd(f"data/{output_dir}")
        Path(abs_dir).mkdir(parents=True, exist_ok=True)
        last_path = os.path.join(abs_dir, "LAST.json")
        timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")
        with open(last_path, "w") as datafile:
            json.dump(json_data, datafile)
        with open(timestamp_path, "w") as datafile:
            json.dump(json_data, datafile)

        print(f"data written to {timestamp_path} and {last_path}", end="\r", flush=True)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--scheduler",
        "-s",
        type=str,
        choices=["no_sched", "fcfs", "qoala"],
        required=True,
    )
    parser.add_argument("--num_iterations", "-n", type=int, required=True)
    parser.add_argument("--dump", "-d", action="store_true")
    parser.add_argument("--log", "-l", action="store_true")

    args = parser.parse_args()

    num_iterations = args.num_iterations
    dump = args.dump
    log = args.log

    if args.scheduler == "no_sched":
        run(
            output_dir="no_sched",
            sched_typ=SchedulerType.NO_SCHED,
            num_iterations=num_iterations,
            dump=dump,
            log=log,
        )
    elif args.scheduler == "fcfs":
        run(
            output_dir="fcfs",
            sched_typ=SchedulerType.FCFS,
            num_iterations=num_iterations,
            dump=dump,
            log=log,
        )
    elif args.scheduler == "qoala":
        run(
            output_dir="qoala",
            sched_typ=SchedulerType.QOALA,
            num_iterations=num_iterations,
            dump=dump,
            log=log,
        )



import logging
from typing import Optional, Union

import netsquid as ns


class SimTimeFilter(logging.Filter):
    def filter(self, record):
        record.simtime = f"{ns.sim_time():_}"
        return True


class LogManager:
    STACK_LOGGER = "Stack"
    _STACK_LOGGER_HAS_BEEN_SETUP = False
    TASK_LOGGER = "Scheduler"
    _TASK_LOGGER_HAS_BEEN_SETUP = False

    @classmethod
    def _setup_stack_logger(cls) -> None:
        logger = logging.getLogger(cls.STACK_LOGGER)
        formatter = logging.Formatter(
            "%(levelname)s:%(simtime)s ns:%(name)s:%(message)s"
        )
        syslog = logging.StreamHandler()
        syslog.setFormatter(formatter)
        syslog.addFilter(SimTimeFilter())
        logger.addHandler(syslog)
        logger.propagate = False
        cls._STACK_LOGGER_HAS_BEEN_SETUP = True

    @classmethod
    def _setup_task_logger(cls) -> None:
        logger = logging.getLogger(cls.TASK_LOGGER)
        formatter = logging.Formatter(
            "%(levelname)s:%(simtime)s ns:%(name)s:%(message)s"
        )
        syslog = logging.StreamHandler()
        syslog.setFormatter(formatter)
        syslog.addFilter(SimTimeFilter())
        logger.addHandler(syslog)
        logger.propagate = False
        logger.setLevel(logging.CRITICAL + 1)
        cls._TASK_LOGGER_HAS_BEEN_SETUP = True

    @classmethod
    def get_stack_logger(cls, sub_logger: Optional[str] = None) -> logging.Logger:
        if not cls._STACK_LOGGER_HAS_BEEN_SETUP:
            cls._setup_stack_logger()
        logger = logging.getLogger(cls.STACK_LOGGER)
        if sub_logger is None:
            return logger
        else:
            return logger.getChild(sub_logger)

    @classmethod
    def get_task_logger(cls, sub_logger: Optional[str] = None) -> logging.Logger:
        if not cls._TASK_LOGGER_HAS_BEEN_SETUP:
            cls._setup_task_logger()
        logger = logging.getLogger(cls.TASK_LOGGER)
        if sub_logger is None:
            return logger
        else:
            return logger.getChild(sub_logger)

    @classmethod
    def set_log_level(cls, level: Union[int, str]) -> None:
        logger = cls.get_stack_logger()
        logger.setLevel(level)

    @classmethod
    def get_log_level(cls) -> int:
        return cls.get_stack_logger().level

    @classmethod
    def set_task_log_level(cls, level: Union[int, str]) -> None:
        logger = cls.get_task_logger()
        logger.setLevel(level)

    @classmethod
    def log_to_file(cls, path: str) -> None:
        file_handler = logging.FileHandler(path, mode="w")
        formatter = logging.Formatter(
            "%(levelname)s:%(simtime)s ns:%(name)s:%(message)s"
        )
        file_handler.setFormatter(formatter)
        file_handler.addFilter(SimTimeFilter())
        cls.get_stack_logger().addHandler(file_handler)

    @classmethod
    def log_tasks_to_file(cls, path: str) -> None:
        file_handler = logging.FileHandler(path, mode="w")
        formatter = logging.Formatter(
            "%(levelname)s:%(simtime)s ns:%(name)s:%(message)s"
        )
        file_handler.setFormatter(formatter)
        file_handler.addFilter(SimTimeFilter())
        cls.get_task_logger().addHandler(file_handler)
